package snow;

import java.awt.event.ActionEvent;
import java.util.*;

import javax.swing.*;
import javax.swing.event.*;
import static snow.SWConstants.*;

public class SWGUIModel
{
	static final String PLOW_BOUNDARY_ONLY = "Plowing must start on a boundary.";
	static final String NO_PLOWS_LEFT = "No plows left.";
	static final String NO_STREETS_LEFT = "Maximum number of streets plowed.";
	static final String NO_RETRACE =
    "This street has already been plowed (no retrace allowed).";
	static final String NO_UNDO = "No moves to undo.";
	
	List<ChangeListener> listeners = new ArrayList<ChangeListener>();
	ChangeEvent event;
	boolean batchMode = false;
	boolean firePending = false;
	Action resetAction = new ResetAction();
	Action undoAction = new UndoAction();
		
	int maxPlows;
	int maxTotalStreets;
	boolean boundaryOnly;
	boolean allowRetrace;
	StringBuilder script;

	SWModel model;
	Plow plow;
	ArrayList<Integer> plowCounts;  // Number of streets plowed by each plow.
	String message = null;

  // The block (if any) currently selected for inspecting routes.
	Block selectedBlock;
	
	public SWGUIModel()
	{
		event = new ChangeEvent(this);
		restoreDefaults();
	}

	public void startBatch()
	{
	  batchMode = true;
	}
	
	public void endBatch()
	{
	  batchMode = false;
	  if (firePending)
	  {
	    firePending = false;
	    fireStateChanged();
	  }
	}
	
	public void addChangeListener(ChangeListener cl)
	{
		listeners.add(cl);
	}
	
	protected void fireStateChanged()
	{
	  if (batchMode)
	  {
	    firePending = true;
	    return;
	  }
		for (ChangeListener cl : listeners)
		{
			cl.stateChanged(event);
		}
	}

	void restoreDefaults()
	{
	  startBatch();
	  doReset(INIT_WIDTH, INIT_HEIGHT, INIT_CORNERS, INIT_WALKTHROUGH);
	  restoreWalkThroughDefaults();
	  setMaxPlows(INIT_PLOWS);
	  setMaxTotalStreets(INIT_LENGTH);
	  setBoundaryOnly(INIT_BOUNDARY_ONLY);
	  setAllowRetrace(INIT_ALLOW_RETRACE);
	  endBatch();
	}
	
	void reset()
	{
		doReset();
		fireStateChanged();
	}
	
	void doReset()
	{
		doReset(
      model.width(), model.height(), model.corners(), model.walkThrough());
	}
	
	void doReset(int w, int h, boolean corners, boolean walkThrough)
	{
		model = new SWModel(w, h, corners, walkThrough);
		plow = new Plow();
		plowCounts = new ArrayList<Integer>();
		script = new StringBuilder();
		clearMessage();
	}
	
	public void corners(boolean b)
	{
	  if (b == corners())
	  {
	    return;
	  }
		model.corners(b);
		fireStateChanged();
	}
	
	public int maxPlows()
	{
		return maxPlows;
	}

	public int plowsRemaining()
	{
		return maxPlows == 0 ? Integer.MAX_VALUE : maxPlows - plowsUsed();
	}
	
	public int maxTotalStreets()
	{
		return maxTotalStreets;
	}

	public int totalStreetsRemaining()
	{
		return maxTotalStreets == 0 ? Integer.MAX_VALUE : maxTotalStreets - totalStreetsPlowed();
	}
	
	public boolean boundaryOnly()
	{
		return boundaryOnly;
	}

	public boolean allowRetrace()
	{
		return allowRetrace;
	}

	public boolean walkThrough()
	{
		return model.walkThrough();
	}
	
	public String message()
	{
		return message;
	}

	public void clearMessage()
	{
		message = null;
	}
	
	public void setMaxPlows(int n)
	{
	  clearMessage();
	  if (n == maxPlows())
	  {
	    return;
	  }
		if (n != 0 && n < plowsUsed())
		{
			doReset();
		}
		maxPlows = n;
		fireStateChanged();
	}
	
	void setDimension(int w, int h)
	{
    clearMessage();
	  if (w == width() && h == height())
	  {
	    return;
	  }
		doReset(w, h, corners(), walkThrough());
		fireStateChanged();
	}
	
	public void setMaxTotalStreets(int n)
	{
    clearMessage();
	  if (maxTotalStreets() == n)
	  {
	    return;
	  }
		if (n != 0 && n < totalStreetsPlowed())
		{
			doReset();
		}
		maxTotalStreets = n;
		fireStateChanged();
	}

	public void setBoundaryOnly(boolean b)
	{
    clearMessage();
	  if (b == boundaryOnly())
	  {
	    return;
	  }
		if (!boundaryOnly && b)
		{
			doReset();
		}
		boundaryOnly = b;
		fireStateChanged();
	}

	public void setAllowRetrace(boolean b)
	{
    clearMessage();
	  if (b == allowRetrace)
	  {
	    return;
	  }
		if (allowRetrace && !b)
		{
			doReset();
		}
		allowRetrace = b;
		fireStateChanged();
	}

	public void setWalkThrough(boolean b)
	{
    clearMessage();
	  if (b == walkThrough())
	  {
	    return;
	  }
		model.walkThrough(b);
		fireStateChanged();
	}

	public void setWalkThroughCosts(double[] costs)
	{
    clearMessage();
	  for (int i = 0; i < COSTS; ++i)
	  {
	    model.walkThroughCost(i, costs[i]);	    
	  }
	  fireStateChanged();
	}
	
	public void restoreWalkThroughDefaults()
	{
    clearMessage();
	  model.restoreWalkThroughDefaults();
	  fireStateChanged();
	}
	
	public double walkThroughCost(int costType)
	{
	  return model.walkThroughCost(costType);
	}
	
	public int width()
	{
		return model.width();
	}

	public int height()
	{
		return model.height();
	}

	public boolean corners()
	{
		return model.corners();
	}

	public double[] maxDists(int bx, int by)
	{
		return model.maxDists(bx, by);
	}

	public boolean ns(int ix, int iy)
	{
		return model.ns(ix, iy);
	}

	public boolean ew(int ix, int iy)
	{
		return model.ew(ix, iy);
	}

	public void ns(int ix, int iy, boolean b)
	{
		model.ns(ix, iy, b);
	}

	public void ew(int ix, int iy, boolean b)
	{
		model.ew(ix, iy, b);
	}
	
	public int plowX()
	{
		return plow.ix;
	}

	public int plowY()
	{
		return plow.iy;
	}
	
	public boolean plowing()
	{
		return plow.plowing;
	}

	public String script()
	{
	  return script.toString();
	}
	
	public void appendWhitespace(char w)
	{
	  if (Character.isWhitespace(w))
	  {
	    script.append(w);
	  }
	}
	
	void incrementPlowCount()
	{
	  int i = plowCounts.size() - 1;
	  int n = plowCounts.get(i);
	  plowCounts.set(i, n + 1);
	}
	
	public void plowUp()
	{
    clearMessage();
    if (plow.iy > 0)
    {
      if (plow.plowing)
      {
        boolean alreadyPlowed = ns(plow.ix, plow.iy - 1);
      	if (totalStreetsRemaining() <= 0)
      	{
      		message = NO_STREETS_LEFT;
      	}
      	else if (!allowRetrace() && alreadyPlowed)
      	{
      		message = NO_RETRACE;
      	}
      	else
      	{
          script.append("N");
          if (!alreadyPlowed)
          {
            ns(plow.ix, plow.iy - 1, true);
            incrementPlowCount();            
          }
          --plow.iy;
      	}
      }
      else
      {
        script.append("N");
        --plow.iy;      	
      }
      fireStateChanged();
    }	
	}
	
	public void plowDown()
	{
    clearMessage();
    if (plow.iy < height())
    {
      if (plow.plowing)
      {
        boolean alreadyPlowed = ns(plow.ix, plow.iy);
      	if (totalStreetsRemaining() <= 0)
      	{
      		message = NO_STREETS_LEFT;
      	}
      	else if (!allowRetrace() && alreadyPlowed)
      	{
      		message = NO_RETRACE;
      	}
      	else
      	{
          script.append("S");
          if (!alreadyPlowed)
          {
            ns(plow.ix, plow.iy, true);
            incrementPlowCount();
          }
      		++plow.iy;
      	}
      }
      else
      {
        script.append("S");
        ++plow.iy;      	
      }
      fireStateChanged();
    }	
	}
	
	public void plowLeft()
	{
    clearMessage();
    if (plow.ix > 0)
    {            
      if (plow.plowing)
      {
        boolean alreadyPlowed = ew(plow.ix - 1, plow.iy);
      	if (totalStreetsRemaining() <= 0)
      	{
      		message = NO_STREETS_LEFT;
      	}
      	else if (!allowRetrace() && alreadyPlowed)
      	{
      		message = NO_RETRACE;
      	}
      	else
      	{
          script.append("W");
          if (!alreadyPlowed)
          {
            ew(plow.ix - 1, plow.iy, true);
            incrementPlowCount();            
          }
          --plow.ix;
      	}
      }
      else
      {
        script.append("W");
        --plow.ix;      	
      }
      fireStateChanged();
    }	
	}
	
	public void plowRight()
	{
    clearMessage();
    if (plow.ix < width())
    {
      if (plow.plowing)
      {
        boolean alreadyPlowed = ew(plow.ix, plow.iy);
      	if (totalStreetsRemaining() <= 0)
      	{
      		message = NO_STREETS_LEFT;
      	}
      	else if (!allowRetrace() && alreadyPlowed)
      	{
      		message = NO_RETRACE;
      	}
      	else
      	{
      	  script.append("E");
          if (!alreadyPlowed)
          {
            ew(plow.ix, plow.iy, true);
            incrementPlowCount();
          }
          ++plow.ix;
      	}
      }
      else
      {
        script.append("E");
        ++plow.ix;      	
      }
      fireStateChanged();
    }
	}

	public void setPlowing(boolean b)
	{
    clearMessage();
		if (b == plow.plowing)
		{
			return;
		}
		if (b && boundaryOnly && !plow.onBoundary())
		{
			message = PLOW_BOUNDARY_ONLY;
		}
		else if (b && plowsRemaining() == 0)
		{
			message = NO_PLOWS_LEFT;
		}
		else
		{
		  script.append(b ? START_PLOW_STRING : END_PLOW_STRING);
			plow.plowing = b;
		  if (b)
			{
		    plowCounts.add(0);
			}
		}
		fireStateChanged();
	}

	public int plowsUsed()
	{
		return plowCounts.size();
	}

	public int totalStreetsPlowed()
	{
		return model.totalStreetsPlowed();
	}

	public double score()
	{
		return model.score();
	}

	public int maxStreets()
	{
	  int ret = 0;
	  for (int i : plowCounts)
	  {
	    if (i > ret)
	    {
	      ret = i;
	    }
	  }
	  return ret;
	}
	
  public String streets()
  {
    return plowCounts.toString();
  }
  
	/**
	 * @return the block currently selected for inspection, if any
	 */
	public Block selectedBlock()
	{
	  return selectedBlock;
	}
	
	/**
	 * @return whether a block is currently selected for inspection
	 */
	public boolean isBlockSelected()
	{
	  return selectedBlock != null;
	}
	
	/**
   * Return whether the specified block is the the block currently selected
   * for inspection.
	 * @param bx x-coordinate of block
	 * @param by y-coordinate of block
	 * @return whether this block is being inspected
	 */
	public boolean isBlockSelected(int bx, int by)
	{
	  return
      isBlockSelected() && selectedBlock.bx == bx && selectedBlock.by == by; 
	}
	
	/**
	 * Select the specified block for inspection.
	 * @param bx x-coordinate of the block
	 * @param by y-coordinate of the block
	 */
	public void selectBlock(int bx, int by)
	{
	  if (isBlockSelected(bx, by))
	  {
	    selectedBlock = null;
	  }
	  else
	  {
	    selectedBlock = new Block(bx, by);
	  }
	  fireStateChanged();
	}
	
	/**
	 * Get the route to the currently selected block
	 * from the adjacent block in the given direction.
	 * @param dir the direction of the adjacent block from the selected block
	 * @return a list of vertices comprising the shortest path,
	 * including the start and end blocks
	 */
	public List<Vertex> getSelectedRoute(int dir)
	{
	  return isBlockSelected() ? model.getRoute(selectedBlock, dir) : null;
	}
	
	String processScript(String script)
	{
	  startBatch();
	  reset();
	  String error = null;
	  try
	  {
	    doProcessScript(script);
	  }
	  catch (SWParseException swpx)
	  {
	    error = swpx.getMessage();
	  }
	  finally
	  {
	    endBatch();
	  }
	  if (error != null)
	  {
	    JOptionPane.showMessageDialog(
	      null, error, "Error", JOptionPane.ERROR_MESSAGE);
	  }
	  return error;
	}

	void doProcessScript(String script) throws SWParseException
  {
    for (int i = 0; i < script.length(); ++i)
    {
      String c = script.substring(i, i + 1);
      if (Character.isWhitespace(c.charAt(0)))
      {
        appendWhitespace(c.charAt(0));
      }
      else if (START_PLOW_STRING.equalsIgnoreCase(c))
      {
        setPlowing(true);
      }
      else if (END_PLOW_STRING.equalsIgnoreCase(c))
      {
        setPlowing(false);
      }
      else if ("N".equalsIgnoreCase(c))
      {
        plowUp();
      }
      else if ("E".equalsIgnoreCase(c))
      {
        plowRight();
      }
      else if ("S".equalsIgnoreCase(c))
      {
        plowDown();
      }
      else if ("W".equalsIgnoreCase(c))
      {
        plowLeft();
      }
      else
      {
        throw new SWParseException("Unknown action \"" + c + "\"");
      }

      if (message() != null)
      {
        String error =
          "The following error was produced during execution:\n\n    " +
          message() + "\n\n" +
          "Execution was stopped when the error occurred.";
        throw new SWParseException(error);
      }
    }
  }
  
	void undo()
	{
	  String curScript = script();
	  int i = curScript.length() - 1;
	  for ( ; i >= 0; --i)
	  {
	    if (!Character.isWhitespace(curScript.charAt(i)))
	    {
	      break;
	    }
	  }
	  if (i < 0)
	  {
	    message = NO_UNDO;
	  }
	  else
	  {
	    String newScript = curScript.substring(0, i);
	    processScript(newScript);
	  }
	  fireStateChanged();
	}
	
	class UndoAction extends AbstractAction
	{
	  UndoAction()
	  {
	    super("Undo Last Move");
	  }
	  
	  public void actionPerformed(ActionEvent e)
	  {
	    undo();
	  }
	}
	
	class ResetAction extends AbstractAction
	{
	  ResetAction()
	  {
	    super("Reset");
	  }
	  
    public void actionPerformed(ActionEvent e)
    {
      reset();
    }
	}
	
	/**
	 * Class representing a plow.  Stores the current location
	 * (intersection) of the plow and whether the plow
	 * is up or down.
	 */
	private class Plow
	{
	  int ix = 0;
	  int iy = 0;
	  boolean plowing = false;
	  
	  boolean onBoundary()
	  {
	  	return ix == 0 || ix == width() || iy == 0 || iy == height();
	  }
	}

  static String formatInt(int n)
  {
    return n == Integer.MAX_VALUE ? "-" : Integer.toString(n);
  }

  static String formatDouble(double d)
  {
    return d == Double.POSITIVE_INFINITY ? "-" : String.format("%.1f", d);
  }

  static String formatBoolean(boolean b)
  {
    return b ? "YES" : "NO";
  }
}
